from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.graph import Graph
from pybbn.graph.node import BbnNode
from pybbn.graph.variable import Variable


class Dag(Graph):
    """
    Directed acyclic graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        Graph.__init__(self)

    def get_parents(self, id):
        """
        Gets the parent IDs of the specified node.
        :param id: Node id.
        :return: Array of parent ids.
        """
        return [x for x in self.map if id in self.map[x]]

    def get_children(self, node_id):
        """
        Gets the children IDs of the specified node.
        :param node_id: Node id.
        :return: Array of children ids.
        """
        return [x for x in self.map[node_id]]

    def __shouldadd__(self, edge):
        """
        Checks if the specified directed edge should be added.
        :param edge: Directed edge.
        :return: A boolean indicating if the edge should be added.
        """
        if EdgeType.DIRECTED != edge.type:
            return False

        parent = edge.i
        child = edge.j

        if parent.id == child.id:
            return False

        if child.id not in self.map[parent.id] and parent.id not in self.map[child.id]:
            if not PathDetector(self, child.id, parent.id).exists():
                return True

        return False

    def edge_exists(self, id1, id2):
        """
        Checks if a directed edge exists between the specified id. e.g. id1 -> id2
        :param id1: Node id.
        :param id2: Node id.
        :return: A boolean indicating if a directed edge id1 -> id2 exists.
        """
        if id2 in self.map[id1] and id1 not in self.map[id2]:
            return True
        return False


class Bbn(Dag):
    """
    BBN.
    """

    def __init__(self):
        """Ctor."""
        Dag.__init__(self)

    def __shouldadd__(self, edge):
        """
        Checks if the specified directed edge should be added.
        :param edge: Directed edge.
        :return: A boolean indicating if the directed edge should be added.
        """
        if isinstance(edge.i, BbnNode) and isinstance(edge.j, BbnNode):
            return True
        return Dag.__shouldadd__(edge)


class PathDetector(object):
    """
    Detects path between two nodes.
    """

    def __init__(self, graph, start, stop):
        """
        :param graph: DAG.
        :param start: Start node id.
        :param stop: Stop node id.
        """
        self.graph = graph
        self.start = start
        self.stop = stop
        self.seen = set()

    def exists(self):
        """
        Checks if a path exists.
        :return: True if a path exists, otherwise, false.
        """
        if self.start == self.stop:
            return True
        else:
            return self.__find__(self.start)

    def __find__(self, i):
        """
        Checks if a path exists from the specified node to the stop node.
        :param i: Node id.
        :return: True if a path exists, otherwise, false.
        """
        children = self.graph.get_children(i)
        if self.stop in children:
            return True

        self.seen.add(i)
        for child in children:
            if child not in self.seen and self.__find__(child):
                return True

        return False


class BbnUtil(object):
    """
    BBN utility.
    """

    @staticmethod
    def get_huang_graph():
        """
        Gets the Huang reference BBN graph.
        :return: BBN.
        """
        a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
        b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
        c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
        d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])
        e = BbnNode(Variable(4, 'e', ['on', 'off']), [0.3, 0.7, 0.6, 0.4])
        f = BbnNode(Variable(5, 'f', ['on', 'off']), [0.01, 0.99, 0.01, 0.99, 0.01, 0.99, 0.99, 0.01])
        g = BbnNode(Variable(6, 'g', ['on', 'off']), [0.8, 0.2, 0.1, 0.9])
        h = BbnNode(Variable(7, 'h', ['on', 'off']), [0.05, 0.95, 0.95, 0.05, 0.95, 0.05, 0.95, 0.05])

        bbn = Bbn() \
            .add_node(a) \
            .add_node(b) \
            .add_node(c) \
            .add_node(d) \
            .add_node(e) \
            .add_node(f) \
            .add_node(g) \
            .add_node(h) \
            .add_edge(Edge(a, b, EdgeType.DIRECTED)) \
            .add_edge(Edge(a, c, EdgeType.DIRECTED)) \
            .add_edge(Edge(b, d, EdgeType.DIRECTED)) \
            .add_edge(Edge(c, e, EdgeType.DIRECTED)) \
            .add_edge(Edge(d, f, EdgeType.DIRECTED)) \
            .add_edge(Edge(e, f, EdgeType.DIRECTED)) \
            .add_edge(Edge(c, g, EdgeType.DIRECTED)) \
            .add_edge(Edge(e, h, EdgeType.DIRECTED)) \
            .add_edge(Edge(g, h, EdgeType.DIRECTED))

        return bbn



from nose import with_setup

from pybbn.graph.node import BbnNode, Clique
from pybbn.graph.variable import Variable


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_bbn_node_creation():
    """
    Tests BBN node creation.
    :return: None.
    """
    a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
    b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
    c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
    d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])
    e = BbnNode(Variable(4, 'e', ['on', 'off']), [0.3, 0.7, 0.6, 0.4])
    f = BbnNode(Variable(5, 'f', ['on', 'off']), [0.01, 0.99, 0.01, 0.99, 0.01, 0.99, 0.99, 0.01])
    g = BbnNode(Variable(6, 'g', ['on', 'off']), [0.8, 0.2, 0.1, 0.9])
    h = BbnNode(Variable(7, 'h', ['on', 'off']), [0.05, 0.95, 0.95, 0.05, 0.95, 0.05, 0.95, 0.05])

    assert a.id == 0
    assert b.id == 1
    assert c.id == 2
    assert d.id == 3
    assert e.id == 4
    assert f.id == 5
    assert g.id == 6
    assert h.id == 7

    assert a.get_weight() == 2
    assert b.get_weight() == 2
    assert c.get_weight() == 2
    assert d.get_weight() == 2
    assert e.get_weight() == 2
    assert f.get_weight() == 2
    assert g.get_weight() == 2
    assert h.get_weight() == 2


@with_setup(setup, teardown)
def test_clique_creation():
    """
    Tests clique creation.
    :return: None.
    """
    a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
    b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
    c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
    clique = Clique([a, b, c])

    assert clique.id == 45718137
    assert len(clique.nodes) == 3
    assert clique.get_weight() == 8
    assert clique.contains(0)
    assert clique.contains(1)
    assert clique.contains(2)
    assert clique.contains(3) == 0


@with_setup(setup, teardown)
def test_clique_is_superset():
    """
    Tests if clique is a superset of another clique.
    :return: None.
    """
    a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
    b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
    c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
    d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])

    abc = Clique([a, b, c])
    abc2 = Clique([a, b, c])
    ab = Clique([a, b])
    ac = Clique([a, c])
    bc = Clique([b, c])
    cd = Clique([c, d])

    assert abc.is_superset(Clique([a])) == 1
    assert abc.is_superset(Clique([b])) == 1
    assert abc.is_superset(Clique([c])) == 1
    assert abc.is_superset(Clique([d])) == 0
    assert abc.is_superset(abc2) == 1
    assert abc.is_superset(ab) == 1
    assert abc.is_superset(ac) == 1
    assert abc.is_superset(bc) == 1
    assert abc.is_superset(cd) == 0


@with_setup(setup, teardown)
def test_sep_set_creation():
    """
    Tests separation set creation.
    :return: None.
    """
    a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
    b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
    c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
    d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])
    abc = Clique([a, b, c])
    bcd = Clique([b, c, d])
    sepset = abc.get_sep_set(bcd)

    assert len(sepset.nodes) == 2

    names = [node.variable.name for node in sepset.nodes]
    assert 'b' in names
    assert 'c' in names

    assert sepset.get_cost() == 16
    assert sepset.get_mass() == 2


@with_setup(setup, teardown)
def test_str():
    """
    Tests str function.
    :return: None.
    """
    a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
    b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
    c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
    d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])
    abc = Clique([a, b, c])
    bcd = Clique([b, c, d])
    sepset = abc.get_sep_set(bcd)

    print(a)
    print(b)
    print(c)
    print(d)
    print(abc)
    print(bcd)
    print(sepset)

    assert a.__str__() == '0|a|on,off'
    assert b.__str__() == '1|b|on,off'
    assert c.__str__() == '2|c|on,off'
    assert d.__str__() == '3|d|on,off'
    assert abc.__str__() == '(a,b,c)'
    assert bcd.__str__() == '(b,c,d)'
    assert sepset.__str__() == '|b,c|'



import json

import networkx as nx
import numpy as np
from networkx.algorithms.dag import is_directed_acyclic_graph
from networkx.algorithms.shortest_paths.generic import shortest_path
from scipy.stats import dirichlet

from pybbn.graph.dag import Bbn
from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.node import BbnNode
from pybbn.graph.variable import Variable


def __get_simple_ordered_tree__(n):
    """
    Generates a simple-ordered tree. The tree is just a
    directed acyclic graph of n nodes with the structure
    0 --> 1 --> .... --> n.
    :param n: Number of nodes.
    :return: A directed graph.
    """
    g = nx.DiGraph()

    for i in range(n):
        g.add_node(i)

    for i in range(n - 1):
        g.add_edges_from([(i, i + 1, {})])
    return g


def __convert_to_undirected_graph__(g):
    """
    Converts a directed acyclic graph (DAG) to an undirected graph.
    We need to convert a DAG to an undirected one to use
    some API calls to operate over the undirected graph. For example,
    in checking for connectedness of a graph, the API has a method
    to check for connectedness of an undirected graph, but not a
    DAG.
    :param g: Graph.
    :return: An undirected graph.
    """
    u = nx.Graph()
    for n in g.nodes:
        u.add_node(n)
    for e in g.edges:
        u.add_edges_from([(e[0], e[1], {})])
    return u


def __is_connected__(g):
    """
    Checks if a the directed acyclic graph is connected.
    :return: A boolean indicating if the graph is connected.
    """
    u = __convert_to_undirected_graph__(g)
    return nx.is_connected(u)


def __get_random_node_pair__(n):
    """
    Randomly generates a pair of nodes.
    :param n: Number of nodes.
    :return: A tuple of random nodes.
    """
    i = np.random.randint(0, n)
    j = i
    while j == i:
        j = np.random.randint(0, n)
    return i, j


def __edge_exists__(i, j, g):
    """
    Checks if the edge i --> j exists in the graph, g.
    :param i: Index of a node.
    :param j: Index of a node.
    :param g: Graph.
    :return: A boolean indicating if j is a successor of i.
    """
    return j in list(g.successors(i))


def __del_edge__(i, j, g):
    """
    Deletes the edge i --> j in the graph, g. The edge is only
    deleted if this removal does NOT cause the graph to be
    disconnected.
    :param i: Index of a node.
    :param j: Index of a node.
    :param g: Graph.
    :return: None
    """
    if g.has_edge(i, j) is True:
        g.remove_edge(i, j)

        if __is_connected__(g) is False:
            g.add_edges_from([(i, j, {})])


def __add_edge__(i, j, g):
    """
    Adds an edge i --> j to the graph, g. The edge is only
    added if this addition does NOT cause the graph to have
    cycles.
    :param i: Index of a node.
    :param j: Index of a node.
    :param g: Graph.
    :return: None
    """
    g.add_edges_from([(i, j, {})])
    if is_directed_acyclic_graph(g) is False:
        g.remove_edge(i, j)


def __find_predecessor__(i, j, g):
    """
    Finds a predecessor, k, in the path between two nodes, i and j,
    in the graph, g. We assume g is connected, and there is a
    path between i and j (ignoring the direction of the edges).
    We want to find a k, that is a parent of j, that is in
    the path between i and j. In some cases, we may not find
    such a k.
    :param i: Index of node.
    :param j: Index of node.
    :param g: Graph.
    :return: Returns predecessor, if any, or None.
    """
    parents = list(g.predecessors(j))
    u = __convert_to_undirected_graph__(g)
    for pa in parents:
        try:
            shortest_path(u, pa, i)
            return pa
        except nx.NetworkXNoPath:
            pass
    return None


def __generate_multi_connected_structure__(n, max_iter=10):
    """
    Generates a multi-connected directed acyclic graph.
    :param n: Number of nodes.
    :param max_iter: Maximum iterations.
    :return: Graph structure (networkx).
    """
    g = __get_simple_ordered_tree__(n)
    for it in range(max_iter):
        i, j = __get_random_node_pair__(n)
        if g.has_edge(i, j) is True:
            __del_edge__(i, j, g)
        else:
            __add_edge__(i, j, g)
    return g


def __generate_singly_structure__(n, max_iter=10):
    """
    Generates a singly-connected directed acyclic graph.
    :param n: Number of nodes.
    :param max_iter: Maximum iterations.
    :return: Graph structure (networkx).
    """
    g = __get_simple_ordered_tree__(n)

    for it in range(max_iter):
        i, j = __get_random_node_pair__(n)
        if g.has_edge(i, j) is True or g.has_edge(j, i) is True:
            pass
        else:
            p = np.random.random()
            k = __find_predecessor__(i, j, g)
            if k is not None:
                g.remove_edge(k, j)
                if p < 0.5:
                    g.add_edges_from([(j, i, {})])
                else:
                    g.add_edges_from([(i, j, {})])

                if __is_connected__(g) is False:
                    g.add_edges_from([(k, j, {})])

                    if p < 0.5:
                        g.remove_edge(j, i)
                    else:
                        g.remove_edge(i, j)
    return g


def __generate_num_values__(n, max_values=2):
    """
    For each node, i, in the nodes, n, determine the number of values
    the node (or equivalently, variable) has. Every node/variable in a
    Bayesian Network should have 2 or more values. This generates
    the number of values each variable will have. Each number will be
    sampled uniformly.
    :param n: Number of nodes.
    :param max_values: Maximum number of values for a node.
    :return: Array of number of values for each node.
    """
    return np.array([max(np.random.randint(0, max_values) + 1, 2) for _ in range(n)])


def __generate_alphas__(n, max_alpha=10):
    """
    Generate random number for the alpha's (the hyperparameters).
    Each number will be in the range [1, max_alpha]. Each number will
    be sampled uniformly.
    :param n: Number of alpha's to sample.
    :param max_alpha: Maximum alpha value.
    :return: A list of alpha's.
    """
    return [np.random.randint(1, max_alpha + 1) for i in range(n)]


def __sample_dirichlet__(n, max_alpha=10):
    """
    Samples from the Dirichlet distribution to a produce
    a probability vector of length n. The sum of each probability
    in the probability vector should sum to 1.
    :param n: Number of alpha's to sample.
    :param max_alpha: The maximum alpha.
    :return: Array of Dirichlet distributed values.
    """
    return np.array(dirichlet.rvs(__generate_alphas__(n, max_alpha))[0])


def __get_num_parent_instantiations__(parents, num_values):
    """
    Gets the number of parent instantiations.
    :param parents: List of parent indices.
    :param num_values: List of the number of values per node.
    :return: Number of parent instantiations.
    """
    num_pa_instantiations = 1
    for pa in parents:
        num_pa_values = num_values[pa]
        num_pa_instantiations *= num_pa_values
    return num_pa_instantiations


def __generate_dirichlet_parameters__(i, parents, num_values, max_alpha=10):
    """
    Randomly and uniformly generate parameters for a node i. A matrix
    of parameters will be returned. The matrix will represent the
    condtional probability table of the node i. The matrix will have
    the dimensions m (rows) by n (columns), m x n, where m is the
    product of the domain sizes of the parents, and n is the domain
    size of the node. The domain size is just the number of values
    that a node (variable) has, which should always be greater than
    or equal to 2.
    :param i: The index of the node for which parameters are being generated.
    :param parents: The indices of the parents of the node.
    :param num_values: The number of values desired.
    :param max_alpha: The maximum alpha per value.
    :return: A conditional probability table (CPT) that specifies the local probability model for the node.
    """
    num_pa_instantiations = __get_num_parent_instantiations__(parents, num_values)

    n = num_values[i]
    cpt = []
    for pa_instantiation in range(num_pa_instantiations):
        probs = __sample_dirichlet__(n, max_alpha)
        cpt.append(probs)
    return np.array(cpt)


def __generate_parameters__(g, max_values=2, max_alpha=10):
    """
    Generates parameters for each node in the graph, g.
    A dictionary indexed by the node's id will give its
    (sampled) parameters and its parents.
    :param max_values: Maximum values per node.
    :param max_alpha: Maximum alpha per value (hyperparameters).
    :return: Parameters.
    """
    num_nodes = len(list(g.nodes))
    num_values = __generate_num_values__(num_nodes, max_values)
    g_params = {}
    for i in g.nodes:
        parents = list(g.predecessors(i))
        params = __generate_dirichlet_parameters__(i, parents, num_values, max_alpha)
        g_params[i] = {
            'parents': parents,
            'params': params,
            'shape': [__get_num_parent_instantiations__(parents, num_values), num_values[i]]
        }
    return g_params


def to_json(g, params, pretty=False):
    """
    Serializes the graph to JSON.
    :param g: Graph.
    :param params: Parameters.
    :param pretty: Pretty-print serialization flag.
    :return: None.
    """
    j = {
        'nodes': list(g.nodes),
        'edges': [{'pa': e[0], 'ch': e[1]} for e in g.edges],
        'parameters': [{
                'node': k,
                'params': list(v['params'].flatten()),
                'shape': v['shape']} for k, v in params.items()
        ]}

    return json.dumps(j, indent=2, sort_keys=False) if pretty is True else json.dumps(j)


def generate_multi_bbn(n, max_iter=10, max_values=2, max_alpha=10):
    """
    Generates structure and parameters for a multi-connected BBN.
    :param n: Number of nodes.
    :param max_iter: Maximum iterations.
    :param max_values: Maximum values per node.
    :param max_alpha: Maximum alpha per value (hyperparameters).
    :return: A tuple of structure and parameters.
    """
    g = __generate_multi_connected_structure__(n, max_iter)
    p = __generate_parameters__(g, max_values, max_alpha)
    return g, p


def generate_singly_bbn(n, max_iter=10, max_values=2, max_alpha=10):
    """
    Generates structure and parameters for a singly-connected BBN.
    :param n: Number of nodes.
    :param max_iter: Maximum iterations.
    :param max_values: Maximum values per node.
    :param max_alpha: Maximum alpha per value (hyperparameters).
    :return: A tuple of structure and parameters.
    """
    g = __generate_singly_structure__(n, max_iter)
    p = __generate_parameters__(g, max_values, max_alpha)
    return g, p


def convert_for_exact_inference(g, p):
    """
    Converts the graph and parameters to a BBN.
    :param g: Directed acyclic graph (DAG in the form of networkx).
    :param p: Parameters.
    :return: BBN.
    """
    bbn = Bbn()

    bbn_nodes = {}

    for node in g.nodes:
        id = node
        params = p[id]['params'].flatten()
        states = ['state{}'.format(state) for state in range(p[id]['shape'][1])]
        v = Variable(id, str(id), states)
        n = BbnNode(v, params)
        bbn.add_node(n)
        bbn_nodes[id] = n

    for e in g.edges:
        pa = bbn_nodes[e[0]]
        ch = bbn_nodes[e[1]]
        bbn.add_edge(Edge(pa, ch, EdgeType.DIRECTED))

    return bbn


def convert_for_drawing(bbn):
    """
    Converts a BBN to a networkx graph for drawing.
    :param bbn: BBN.
    :return: Directed acyclic graph.
    """
    g = nx.DiGraph()

    for k, v in bbn.nodes.iteritems():
        g.add_node(v.id)

    for k, e in bbn.edges.iteritems():
        pa = e.i.id
        ch = e.j.id
        g.add_edges_from([(pa, ch, {})])

    return g



from pybbn.graph.jointree import JoinTreeListener
from pybbn.pptc.initializer import Initializer
from pybbn.pptc.moralizer import Moralizer
from pybbn.pptc.potentialinitializer import PotentialInitializer
from pybbn.pptc.propagator import Propagator
from pybbn.pptc.transformer import Transformer
from pybbn.pptc.triangulator import Triangulator


class InferenceController(JoinTreeListener):
    """
    Inference controller.
    """

    @staticmethod
    def apply(bbn):
        """
        Sets up the specified BBN for probability propagation in tree clusters (PPTC).
        :param bbn: BBN graph.
        :return: Join tree.
        """
        PotentialInitializer.init(bbn)

        ug = Moralizer.moralize(bbn)
        cliques = Triangulator.triangulate(ug)
        join_tree = Transformer.transform(cliques)

        Initializer.initialize(join_tree)
        Propagator.propagate(join_tree)

        join_tree.set_listener(InferenceController())

        return join_tree

    def evidence_retracted(self, join_tree):
        """
        Evidence is retracted.
        :param join_tree: Join tree.
        """
        Initializer.initialize(join_tree)
        Propagator.propagate(join_tree)

    def evidence_updated(self, join_tree):
        """
        Evidence is updated.
        :param join_tree: Join tree.
        """
        Propagator.propagate(join_tree)



from pybbn.graph.graph import Graph


class Pdag(Graph):
    """
    Partially directed acyclic graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        Graph.__init__(self)

    def get_parents(self, id):
        """
        Gets the parent of the specified node id.
        :param id: Node id.
        :return: Array of parent ids.
        """
        return [x for x in self.map if id in self.map[x]]

    def get_out_nodes(self, id):
        """
        Gets all the out nodes for the node with the specified id. Out nodes are all connected ndoes that are
        not parents (do not have a directed arc into the specified node).
        :param id: Node id.
        :return: Array of out node ids.
        """
        parents = self.get_parents(id)
        neighbors = self.get_neighbors(id)
        out_nodes = [neighbor_id for neighbor_id in neighbors if neighbor_id not in parents]
        return out_nodes

    def __shouldadd__(self, edge):
        """
        Checks if the specified edge should be added.
        :param edge: Edge (could be directed or undirected).
        :return: A boolean indicating if the edge should be added.
        """
        parent = edge.i
        child = edge.j

        if parent.id == child.id:
            return False

        if child.id not in self.map[parent.id] and parent.id not in self.map[child.id]:
            if not PathDetector(self, child.id, parent.id).exists():
                return True

        return False

    def edge_exists(self, id1, id2):
        """
        Checks if the specified edge id1 -- id2 exists.
        :param id1: Node id.
        :param id2: Node id.
        :return: A boolean indicating if the edge exists.
        """
        if id2 in self.map[id1] or id1 in self.map[id2]:
            return True
        return False

    def directed_edge_exists(self, id1, id2):
        """
        Checks if the specified edge id1 -> id2 exists.
        :param id1: Node id.
        :param id2: Node id.
        :return: A boolean indicating if the edge exists.
        """
        if id2 in self.map[id1] and id1 not in self.map[id2]:
            return True
        return False


class PathDetector(object):
    """
    Detects path between two nodes.
    """

    def __init__(self, graph, start, stop):
        """
        Ctor.
        :param graph: Pdag.
        :param start: Start node id.
        :param stop: Stop node id.
        """
        self.graph = graph
        self.start = start
        self.stop = stop
        self.seen = set()

    def exists(self):
        """
        Checks if a path exists.
        :return: True if a path exists, otherwise, false.
        """
        if self.start == self.stop:
            return True
        else:
            return self.__find__(self.start)

    def __find__(self, i):
        """
        Checks if a path exists from the specified node to the stop node.
        :param i: Node id.
        :return: True if a path exists, otherwise, false.
        """
        out_nodes = self.graph.get_out_nodes(i)
        if self.stop in out_nodes:
            return True

        self.seen.add(i)
        for out_node in out_nodes:
            if out_node not in self.seen and self.__find__(out_node):
                return True

        return False



from nose import with_setup

from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.node import Node
from pybbn.graph.pdag import Pdag


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_pdag_creation():
    """
    Tests PDAG creation.
    :return:
    """
    n0 = Node(0)
    n1 = Node(1)
    n2 = Node(2)
    e0 = Edge(n0, n1, EdgeType.DIRECTED)
    e1 = Edge(n1, n2, EdgeType.DIRECTED)
    e2 = Edge(n2, n0, EdgeType.DIRECTED)

    g = Pdag()
    g.add_node(n0)
    g.add_node(n1)
    g.add_edge(e0)
    g.add_edge(e1)
    g.add_edge(e2)

    print(g)

    assert len(g.get_nodes()) == 3
    assert len(g.get_edges()) == 2

    assert len(list(g.get_neighbors(0))) == 1
    assert len(list(g.get_neighbors(1))) == 2
    assert len(list(g.get_neighbors(2))) == 1

    assert 1 in g.get_neighbors(0)
    assert 0 in g.get_neighbors(1)
    assert 2 in g.get_neighbors(1)
    assert 1 in g.get_neighbors(2)

    assert g.edge_exists(0, 1) == 1
    assert g.edge_exists(1, 2) == 1
    assert g.edge_exists(0, 2) == 0

    assert g.directed_edge_exists(0, 1) == 1
    assert g.directed_edge_exists(1, 2) == 1
    assert g.directed_edge_exists(0, 2) == 0

    assert len(g.get_parents(0)) == 0
    assert len(g.get_parents(1)) == 1
    assert len(g.get_parents(2)) == 1

    assert 0 in g.get_parents(1)
    assert 1 in g.get_parents(2)

    assert len(g.get_out_nodes(0)) == 1
    assert len(g.get_out_nodes(1)) == 1
    assert len(g.get_out_nodes(2)) == 0

    assert 1 in g.get_out_nodes(0)
    assert 2 in g.get_out_nodes(1)



import math

import networkx as nx
import numpy as np
from networkx.algorithms.dag import topological_sort, is_directed_acyclic_graph

from pybbn.lg.gaussian import rnorm, dnorm, dcmvnorm, RandCondMvn


class Dag(object):
    """
    Directed acyclic graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.g = nx.DiGraph()

    def add_node(self, node_id, metadata={}):
        """
        Adds a node.
        :param node_id: Node ID.
        :param metadata: Metadata (JSON-like).
        :return: DAG.
        """
        self.g.add_node(node_id, metadata=metadata)
        return self

    def add_edge(self, parent_id, child_id, metadata={}):
        """
        Adds an edge.
        :param parent_id: Parent node ID.
        :param child_id: Child node ID.
        :param metadata: Metadata (JSON-like).
        :return: DAG.
        """
        self.g.add_edges_from([(parent_id, child_id, metadata)])
        if is_directed_acyclic_graph(self.g) is False:
            self.g.remove_edge(parent_id, child_id)
            raise ValueError('adding link from parent {} to child {} will create a cycle'.format(parent_id, child_id))
        return self

    def nodes(self):
        """
        Gets a list of all the node IDs.
        :return: List of node IDs.
        """
        return list(self.g.nodes())

    def edges(self):
        """
        Gets a list of all the edges.
        :return: List of edges.
        """
        return list(self.g.edges())

    def number_of_nodes(self):
        """
        Gets the number of nodes.
        :return: Number of nodes.
        """
        return self.g.number_of_nodes()

    def number_of_edges(self):
        """
        Gets the number of edges.
        :return: Number of edges.
        """
        return self.g.number_of_edges()

    def parents(self, node_id):
        """
        Gets the parents for the specified node ID.
        :param node_id: Node ID.
        :return: List of parent node IDs.
        """
        return list(self.g.predecessors(node_id))

    def children(self, node_id):
        """
        Gets the children for the specified node ID.
        :param node_id: Node ID.
        :return: List of children node IDs.
        """
        return list(self.g.successors(node_id))

    def coparents(self, node_id):
        """
        Gets the coparents of the specified node.
        :param node_id:  Node ID.
        :return: List of coparent node IDS.
        """
        copas = set()
        for ch in self.children(node_id):
            pas = self.parents(ch)
            for pa in pas:
                copas.add(pa)
        return list(copas)

    def markov_blanket(self, node_id):
        """
        Gets the Markov blanket of the specified node. The Markov blanket of a node is defined as its parents,
        children, and co-parents.
        :param node_id: Node ID.
        :return: List of node IDs in the Markov blanket.
        """
        blanket = set()

        for pa in self.parents(node_id):
            blanket.add(pa)
        for ch in self.children(node_id):
            blanket.add(ch)
        for copa in self.coparents(node_id):
            blanket.add(copa)

        if node_id in blanket:
            blanket.remove(node_id)
        return sorted(list(blanket))

    def get_sorted_topology(self):
        """
        Gets the sorted topology of the DAG.
        :return: Sorted topology of the DAG.
        """
        return list(topological_sort(self.g))


class Parameters(object):
    """
    Parameters, which are made up of the means and covariance matrix.
    """

    def __init__(self, means, cov):
        """
        Ctor.
        :param means: Means.
        :param cov: Covariance matrix.
        """
        self.means = means
        self.cov = cov

        if len(means) != cov.shape[0]:
            raise ValueError('number of means {} != number of variables {}'.format(len(means), cov.shape[0]))
        if cov.shape[0] != cov.shape[1]:
            raise ValueError('covariance matrix is not square: {} x {}'.format(cov.shape[0], cov.shape[1]))

    def get_mean(self, node_id):
        """
        Gets the mean associated with the node ID.
        :param node_id: Integer node ID.
        :return: Mean.
        """
        return self.means[node_id]

    def get_stdev(self, node_id):
        """
        Gets the standard deviation associated with the node ID.
        :param node_id: Integer node ID.
        :return: Variance.
        """
        return np.sqrt(self.cov[node_id, node_id])


class Bbn(object):
    """
    Bayesian Belief Network.
    """

    def __init__(self, dag, params, max_samples=9000, max_iters=1, mb=False):
        """
        Ctor. Note that the dimensions of the DAG and parameters must match.
        :param dag: DAG.
        :param params: Parameters (means and covariance).
        :param max_samples: Max number of samples to use for approximate inference per iteration. Default is 9,000.
        :param max_iters: Max number of iterations. Default is 1.
        :param mb: A boolean indicating if we are only sampling locally based on Markov Blanket. Default is False.
        """
        self.dag = dag
        self.params = params
        self.evidence = np.array([None for _ in range(dag.number_of_nodes())], dtype=float)
        self.max_samples = max_samples
        self.max_iters = max_iters
        self.mb = mb

        # all of these checks are to make sure the DAG and params align
        # check for 1-to-1 correspondence between nodes and variables
        num_nodes = dag.number_of_nodes()
        num_means = len(params.means)
        dag_nodes = dag.nodes()
        min_node_id = min(dag_nodes)
        max_node_id = max(dag_nodes)

        # num_vars1 = params.cov.shape[0]
        # num_vars2 = params.cov.shape[1]

        if num_nodes != num_means:
            raise ValueError('number of nodes {} != number of means {}'.format(num_nodes, num_means))
        # if num_nodes != num_vars1:
        #     raise ValueError('number of nodes {} != number of variables {}'.format(num_nodes, num_vars1))
        # if num_vars1 != num_vars2:
        #     raise ValueError('covariance is not a square; rows = {}, cols = {}'.format(num_vars1, num_vars2))
        if 0 != min_node_id:
            raise ValueError('Node IDs should start with zero, but it started with {}'.format(min_node_id))
        if num_nodes - 1 != max_node_id:
            raise ValueError('Node IDs should end with number_of_nodes - 1, but it was {}'.format(max_node_id))

        # make sure the node IDs are sequential
        for e, o in zip(range(num_nodes), sorted(dag_nodes)):
            if e != o:
                raise ValueError('Node IDs should be sequential integers but was not: {}'.format(sorted(dag.nodes())))

    def set_evidence(self, node_id, e):
        """
        Sets the evidence for the specified node.
        :param node_id: Node ID.
        :param e: Evidence.
        :return: BBN.
        """
        self.evidence[node_id] = e
        return self

    def get_evidence(self, node_id):
        """
        Gets the evidence for the specified node.
        :param node_id: Node ID.
        :return: Evidence.
        """
        return self.evidence[node_id]

    def get_evidences(self):
        """
        Gets all the evidences.
        :return: Array of evidences.
        """
        return self.evidence

    def clear_evidences(self):
        """
        Clears all the evidences.
        :return: None
        """
        self.evidence = np.array([None for _ in range(self.dag.number_of_nodes())], dtype=float)

    def __has_parents__(self, node_id):
        """
        Checks if the specified node has parents.
        :param node_id: Node ID.
        :return: A boolean indicating if the node has parents.
        """
        return len(self.dag.parents(node_id)) > 0

    def __has_evidence__(self, node_id):
        """
        Checks if the specified node has evidence.
        :param node_id: Node ID.
        :return: A boolean indicating if the node has evidence.
        """
        return False if np.isnan(self.evidence[node_id]) else True

    def do_inference(self):
        """
        Conducts Gibbs sampling.
        :return: The expected state of every variable.
        """

        def get_init_value(bbn, node_id):
            """
            Gets a random initial value for the specified node. If the specified node has evidence set, then that value
            is returned.
            :param bbn: BBN.
            :param node_id: Node id.
            :return: Initial value.
            """
            if bbn.__has_evidence__(node_id):
                return bbn.get_evidence(node_id)
            else:
                m = bbn.params.means[node_id]
                stdev = math.sqrt(bbn.params.cov[node_id, node_id])
                return list(rnorm(1, m, stdev))[0]

        def get_nodes_selector(node_id, num_nodes):
            """
            Gets a selector for all the nodes except the specified node.
            :param node_id: Node id.
            :param num_nodes: Number of nodes.
            :return: Sorted array of node ids.
            """
            return sorted([i for i in range(num_nodes) if i != node_id])

        def get_rcmvnorm_dist(node_id, means, cov):
            """
            Gets a conditional multivariate normal distribution.
            :param node_id: Node id.
            :param means: Means.
            :param cov: Covariance.
            :return: Random conditional multivariate normal distribution.
            """
            num_nodes = cov.shape[0]
            other_nodes = get_nodes_selector(node_id, num_nodes)
            dist = RandCondMvn(means, cov, node_id, other_nodes)
            return dist

        def get_mb_rcmvnorm_dist(dag, node_id, means, cov):
            other_nodes = dag.markov_blanket(node_id)
            dist = RandCondMvn(means, cov, node_id, other_nodes)
            return dist

        def get_rcmvnorm_dists(means, cov):
            """
            Gets a dictionary of rcmvnorm distributions keyed by node ID.
            :param means: Means.
            :param cov: Covariance.
            :return: Dictionary of rcmvnorm distributions.
            """
            dists = {}
            for node_id in range(cov.shape[0]):
                dist = get_rcmvnorm_dist(node_id, means, cov)
                dists[node_id] = dist
            return dists

        def get_mb_rcmvnorm_dists(dag, means, cov):
            """

            :param dag:
            :param means:
            :param cov:
            :return:
            """
            dists = {}
            for node_id in range(cov.shape[0]):
                dist = get_mb_rcmvnorm_dist(dag, node_id, means, cov)
                dists[node_id] = dist
            return dists

        def get_sample(dag, sample, max_samples, dists, mb=False):
            """
            Gets a sample.
            :param dag: Directed acyclic graph.
            :param sample: Initial sample.
            :param max_samples: Max samples.
            :return: Final sample.
            """
            num_nodes = dag.number_of_nodes()

            for i in range(max_samples):
                i_sample = np.copy(sample)

                for node_id in range(num_nodes):
                    if self.__has_evidence__(node_id) is True:
                        x = self.get_evidence(node_id)
                        i_sample[node_id] = x
                    else:
                        other_nodes = dag.markov_blanket(node_id) if mb is True else get_nodes_selector(node_id,
                                                                                                        num_nodes)
                        X = i_sample[other_nodes]
                        i_sample[node_id] = dists[node_id].next(X)

                sample = (sample + i_sample) / 2.0
            return sample

        num_nodes = self.dag.number_of_nodes()
        max_iters = self.max_iters
        outcome = np.zeros((1, num_nodes))
        dists = get_mb_rcmvnorm_dists(self.dag, self.params.means,
                                      self.params.cov) if self.mb is True else get_rcmvnorm_dists(self.params.means,
                                                                                                  self.params.cov)

        for i in range(max_iters):
            sample = np.array([get_init_value(self, node_id) for node_id in range(num_nodes)], dtype=float)
            sample = get_sample(self.dag, sample, self.max_samples, dists, mb=self.mb)
            outcome = outcome + sample

        outcome = outcome / float(max_iters)
        return outcome[0]

    def log_prob(self, data):
        num_nodes = self.dag.number_of_nodes()
        sum = 0.0
        for node_id in range(num_nodes):
            if self.__has_parents__(node_id) is False:
                d = data[:, node_id]
                m = self.params.means[node_id]
                s = math.sqrt(self.params.cov[node_id, node_id])
                logp = np.sum([math.log10(p) for p in dnorm(d, m, s)])
                sum += logp
            else:
                m = self.params.means
                s = self.params.cov
                dep = sorted(self.dag.parents(node_id))
                logp = np.sum([math.log10(p) for p in dcmvnorm(data, m, s, node_id, dep)])
                sum += logp
        return sum



from nose import with_setup

from pybbn.graph.dag import BbnUtil
from pybbn.graph.jointree import EvidenceBuilder
from pybbn.pptc.inferencecontroller import InferenceController


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_inference_controller():
    """
    Tests inference controller.
    :return: None.
    """
    bbn = BbnUtil.get_huang_graph()
    join_tree = InferenceController.apply(bbn)

    print('INIT')
    print_potentials(join_tree)

    ev = EvidenceBuilder() \
        .with_node(join_tree.get_bbn_node(0)) \
        .with_evidence('on', 1.0) \
        .build()

    join_tree.set_observation(ev)

    print('FIRST')
    print_potentials(join_tree)

    # assert 1 == 2


def print_potentials(join_tree):
    """
    Prints the potentials.
    :param join_tree: Join tree.
    :return: None.
    """
    for node in join_tree.get_bbn_nodes():
        potential = join_tree.get_bbn_potential(node)
        print(node)
        print(potential)
        total = sum([entry.value for entry in potential.entries])
        print('total {}'.format(total))
        print('-----')



import itertools


class Potential(object):
    """
    Potential.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.entries = []

    def add_entry(self, entry):
        """
        Adds a PotentialEntry.
        :param entry: PotentialEntry.
        :return: This potential.
        """
        self.entries.append(entry)
        return self

    def get_matching_entries(self, entry):
        """
        Gets all potential entries matching the specified entry.
        :param entry: PotentialEntry.
        :return: Array of matching potential entries.
        """
        return [e for e in self.entries if e.matches(entry)]

    def __str__(self):
        return str.join('\n', [entry.__str__() for entry in self.entries])


class PotentialEntry(object):
    """
    Potential entry.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.entries = dict()
        self.value = 1.0

    def add(self, k, v):
        """
        Adds a node id and its value.
        :param k: Node id.
        :param v: Value.
        :return: This potential entry.
        """
        self.entries[k] = v
        return self

    def matches(self, that):
        """
        Checks if this potential entry matches the specified one. A match is determined with all the keys
        and their associated values in the potential entry passed in matches this one.
        :param that: PotentialEntry.
        :return:
        """
        for k, v in that.entries.items():
            if k not in self.entries or v != self.entries[k]:
                return False
        return True

    def duplicate(self):
        """
        Duplicates this entry.
        :return: PotentialEntry.
        """
        entry = PotentialEntry()
        for k, v in self.entries.items():
            entry.add(k, v)
        entry.value = self.value
        return entry

    def __str__(self):
        arr = ['{}={}'.format(k, v) for k, v in self.entries.items()]
        s = str.join(',', arr)
        return '{}|{}'.format(s, self.value)


class PotentialUtil(object):
    """
    Potential util.
    """

    @staticmethod
    def pass_single_message(join_tree, x, s, y):
        """
        Single message pass from x -- s -- y (from x to s to y).
        :param join_tree: Join tree.
        :param x: Clique.
        :param s: Separation-set.
        :param y: Clique.
        """
        old_sep_set_potential = join_tree.potentials[s.id]
        y_potential = join_tree.potentials[y.id]

        new_sep_set_potential = PotentialUtil.marginalize_for(join_tree, x, s.nodes)
        join_tree.potentials[s.id] = new_sep_set_potential

        PotentialUtil.multiply(y_potential, PotentialUtil.divide(new_sep_set_potential, old_sep_set_potential))

    @staticmethod
    def marginalize_for(join_tree, clique, nodes):
        """
        Marginalizes the specified clique's potential over the specified nodes.
        :param join_tree: Join tree.
        :param clique: Clique.
        :param nodes: List of BBN nodes.
        :return: Potential.
        """
        potential = PotentialUtil.get_potential_from_nodes(nodes)
        clique_potential = join_tree.potentials.get(clique.id)

        for entry in potential.entries:
            matched_entries = clique_potential.get_matching_entries(entry)
            entry.value = sum([matched_entry.value for matched_entry in matched_entries])

        return potential

    @staticmethod
    def normalize(potential):
        """
        Normalizes the potential (make sure they sum to 1.0).
        :param potential: Potential.
        :return: Potential.
        """
        total = sum([entry.value for entry in potential.entries])

        if total != 0.0:
            for entry in potential.entries:
                d = entry.value / total
                entry.value = d

        return potential

    @staticmethod
    def divide(numerator, denominator):
        """
        Divides two potentials.
        :param numerator: Potential.
        :param denominator: Potential.
        :return: Potential.
        """
        potential = Potential()
        for i, entry in enumerate(numerator.entries):
            e = denominator.entries[i]
            d = 0.0 \
                if PotentialUtil.is_zero(entry.value) or PotentialUtil.is_zero(e.value) \
                else (entry.value / e.value)
            new_entry = entry.duplicate()
            new_entry.value = d
            potential.add_entry(new_entry)
        return potential

    @staticmethod
    def is_zero(d):
        """
        Checks if the specified value is 0.0.
        :param d: Value.
        :return: A boolean indicating if the value is zero.
        """
        return 0.0 == d

    @staticmethod
    def multiply(bigger, smaller):
        """
        Multiplies two potentials. Order matters.
        :param bigger: Bigger potential.
        :param smaller: Smaller potential.
        """
        for entry in smaller.entries:
            for e in bigger.get_matching_entries(entry):
                d = e.value * entry.value
                e.value = d

    @staticmethod
    def get_potential(node, parents):
        """
        Gets the potential associated with the specified node and its parents.
        :param node: BBN node.
        :param parents: Parents of the BBN node (that themselves are also BBN nodes).
        :return: Potential.
        """
        potential = PotentialUtil.get_potential_from_nodes(PotentialUtil.merge(node, parents))
        for i in range(len(potential.entries)):
            prob = node.probs[i]
            potential.entries[i].value = prob
        return potential

    @staticmethod
    def get_potential_from_nodes(nodes):
        """
        Gets a potential from a list of BBN nodes.
        :param nodes: Array of BBN nodes.
        :return: Potential.
        """
        lists = [node.variable.values for node in nodes]
        cartesian = PotentialUtil.get_cartesian_product(lists)
        potential = Potential()
        for values in cartesian:
            entry = PotentialEntry()
            for i in range(len(nodes)):
                entry.add(nodes[i].id, values[i])
            potential.add_entry(entry)
        return potential

    @staticmethod
    def get_cartesian_product(lists):
        """
        Gets the cartesian product of a list of lists of values. For example, if the list is
        [ ['on', 'off'], ['on', 'off'] ],
        then the result will be a list of the following.
        * [ 'on', 'on']
        * [ 'on', 'off' ]
        * [ 'off', 'on' ]
        * [ 'off', 'off' ]
        :param lists: List of list of values.
        :return: Cartesian product of values.
        """
        return [i for i in itertools.product(*lists)]

    @staticmethod
    def merge(node, parents):
        """
        Merges the nodes into one array.
        :param node: BBN node.
        :param parents: BBN parent nodes.
        :return: Array of BBN nodes.
        """
        nodes = [parent for parent in parents]
        nodes.append(node)
        return nodes



from nose import with_setup

from pybbn.graph.edge import JtEdge
from pybbn.graph.jointree import JoinTree
from pybbn.graph.node import BbnNode, Clique
from pybbn.graph.variable import Variable


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_jointree_creation():
    """
    Tests join tree creation.
    :return: None.
    """
    n0 = BbnNode(Variable(0, 'n0', ['t', 'f']), [])
    n1 = BbnNode(Variable(1, 'n1', ['t', 'f']), [])
    n2 = BbnNode(Variable(2, 'n2', ['t', 'f']), [])

    clique0 = Clique([n0, n1])
    clique1 = Clique([n1, n2])
    sep_set0 = clique0.get_sep_set(clique1)
    sep_set1 = clique0.get_sep_set(clique1)
    sep_set2 = clique1.get_sep_set(clique0)
    sep_set3 = clique0.get_sep_set(clique0)

    e0 = JtEdge(sep_set0)
    e1 = JtEdge(sep_set1)
    e2 = JtEdge(sep_set2)
    e3 = JtEdge(sep_set3)

    g = JoinTree().add_edge(e0).add_edge(e1).add_edge(e2).add_edge(e3)

    nodes = g.get_nodes()
    edges = g.get_edges()

    assert len(nodes) == 3
    assert len(edges) == 1
    assert len(g.get_flattened_edges()) == 2



from enum import Enum

from pybbn.graph.edge import JtEdge
from pybbn.graph.graph import Ug
from pybbn.graph.node import SepSet
from pybbn.graph.potential import Potential, PotentialEntry, PotentialUtil


class JoinTree(Ug):
    """
    Join tree.
    """

    def __init__(self):
        """
        Ctor.
        """
        Ug.__init__(self)
        self.potentials = dict()
        self.evidences = dict()
        self.listener = None

    def get_bbn_potential(self, node):
        """
        Gets the potential associated with the specified BBN node.
        :param node: BBN node.
        :return: Potential.
        """
        clique = node.metadata['parent.clique']
        return PotentialUtil.normalize(PotentialUtil.marginalize_for(self, clique, [node]))

    def unmark_cliques(self):
        """
        Unmarks the cliques.
        """
        for clique in self.get_cliques():
            clique.unmark()

    def get_bbn_nodes(self):
        """
        Gets all the BBN nodes in this junction tree.
        :return: List of BBN nodes.
        """
        nodes = dict()
        for clique in self.get_cliques():
            for node in clique.nodes:
                nodes[node.id] = node
        return list(nodes.values())

    def get_bbn_node(self, id):
        """
        Gets the BBN node associated with the specified id.
        :param id: Node id.
        :return: BBN node or None if no such node exists.
        """
        for node in self.get_bbn_nodes():
            if id == node.id:
                return node
        return None

    def get_bbn_node_by_name(self, name):
        """
        Gets the BBN node associated with the specified name.
        :param name: Node name.
        :return: BBN node or None if no such node exists.
        """
        for node in self.get_bbn_nodes():
            if name == node.variable.name:
                return node
        return None

    def find_cliques_with_node_and_parents(self, id):
        """
        Finds all cliques in this junction tree having the specified node and its parents.
        :param id: Node id.
        :return: Array of cliques.
        """
        ids = self.__get_parent_ids__(id)
        ids.append(id)

        set1 = set(ids)
        return [clique for clique in self.get_cliques() if set(clique.get_node_ids()).issuperset(set1)]

    def add_potential(self, clique, potential):
        """
        Adds a potential associated with the specified clique.
        :param clique: Clique.
        :param potential: Potential.
        :return: This join tree.
        """
        self.potentials[clique.id] = potential
        return self

    def get_cliques(self):
        """
        Gets all the cliques in this junction tree.
        :return: Array of cliques.
        """
        return [clique for clique in self.get_nodes() if not isinstance(clique, SepSet)]

    def get_sep_sets(self):
        """
        Gets all the separation sets in this junction tree.
        :return: Array of separation sets.
        """
        return [sep_set for sep_set in self.get_nodes() if isinstance(sep_set, SepSet)]

    def add_edge(self, edge):
        """
        Adds an JtEdge.
        :param edge: JtEdge.
        :return: This join tree.
        """
        if not isinstance(edge, JtEdge):
            return self

        sep_set = edge.sep_set
        lhs = edge.i
        rhs = edge.j

        self.add_node(sep_set)
        self.add_node(lhs)
        self.add_node(rhs)

        if sep_set.id not in self.map:
            self.map[sep_set.id] = set()
        if lhs.id not in self.map:
            self.map[lhs.id] = set()
        if rhs.id not in self.map:
            self.map[rhs.id] = set()

        if self.__shouldadd__(edge):
            self.map[lhs.id].add(sep_set.id)
            self.map[rhs.id].add(sep_set.id)

            self.map[sep_set.id].add(lhs.id)
            self.map[sep_set.id].add(rhs.id)

            self.edges[edge.key] = edge
            # lhs_edge = edge.get_lhs_edge()
            # rhs_edge = edge.get_rhs_edge()
            # self.edges[lhs_edge.key] = lhs_edge
            # self.edges[rhs_edge.key] = rhs_edge

        return self

    def get_flattened_edges(self):
        """
        Gets all the edges "flattened" out. Since separation-sets are really hyper-edges, this method breaks
        separation-sets into two edges.
        :return: Array of edges.
        """
        edges = []
        for edge in self.edges.values():
            edges.append(edge.get_lhs_edge())
            edges.append(edge.get_rhs_edge())
        return edges

    def set_listener(self, listener):
        """
        Sets the listener.
        :param listener: JoinTreeListener.
        """
        self.listener = listener

    def get_evidence(self, node, value):
        """
        Gets the evidence associated with the specified BBN node and value.
        :param node: BBN node.
        :param value: Value.
        :return: Potential (the evidence).
        """
        if node.id not in self.evidences:
            self.evidences[node.id] = dict()

        if value not in self.evidences[node.id]:
            entry = PotentialEntry()
            entry.add(node.id, value)
            entry.value = 1.0

            potential = Potential()
            potential.add_entry(entry)

            self.evidences[node.id][value] = potential

        return self.evidences[node.id][value]

    def get_change_type(self, evidences):
        """
        Gets the change type associated with the specified list of evidences.
        :param evidences: List of evidences.
        :return: ChangeType.
        """
        changes = []
        for evidence in evidences:
            node = evidence.node
            potentials = self.evidences[node.id]
            change = evidence.compare(potentials)
            changes.append(change)

        count = len([change_type for change_type in changes if ChangeType.RETRACTION == change_type])
        if count > 0:
            return ChangeType.RETRACTION

        count = len([change_type for change_type in changes if ChangeType.UPDATE == change_type])
        if count > 0:
            return ChangeType.UPDATE

        return ChangeType.NONE

    def get_unobserved_evidence(self, node):
        """
        Gets the unobserved evidences associated with the specified node.
        :param node: BBN node.
        :return: Evidence.
        """
        evidence = Evidence(node, EvidenceType.UNOBSERVE)
        for value in node.variable.values:
            evidence.add_value(value, 1.0)
        return evidence

    def unobserve(self, nodes):
        """
        Unobserves a list of nodeds.
        :param nodes: List of nodes.
        :return: This join tree.
        """
        evidences = [self.get_unobserved_evidence(node) for node in nodes]
        self.update_evidences(evidences)
        return self

    def unobserve_all(self):
        """
        Unobserves all BBN nodes.
        :return: This join tree.
        """
        self.unobserve(self.get_bbn_nodes())
        return self

    def update_evidences(self, evidences):
        """
        Updates this join tree with the list of specified evidence.
        :param evidences: List of evidences.
        :return: This join tree.
        """
        for evidence in evidences:
            evidence.validate()
        change = self.get_change_type(evidences)
        for evidence in evidences:
            node = evidence.node
            potentials = self.evidences[node.id]

            for k, v in evidence.values.items():
                potential = potentials[k]
                potential.entries[0].value = v
        self.__notify_listener__(change)
        return self

    def set_observation(self, evidence):
        """
        Sets a single observation.
        :param evidence: Evidence.
        :return: This join tree.
        """
        if EvidenceType.OBSERVATION != evidence.type:
            return self

        potentials = self.evidences[evidence.node.id]

        pvalues = []
        for v, potential in potentials.items():
            entry = potential.entries[0]
            p = entry.value
            if 1.0 == p:
                pvalues.append(v)

        cvalues = []
        for v, likelihood in evidence.values.items():
            if 1.0 == likelihood:
                cvalues.append(v)

        if 1 == len(pvalues):
            last_value = pvalues[0]
            curr_value = cvalues[0]
            if last_value == curr_value:
                self.unobserve([evidence.node])
            else:
                self.update_evidences([evidence])
        else:
            self.update_evidences([evidence])

        return self

    def __shouldadd__(self, edge):
        """
        Checks if the specified edge should be added.
        :param edge: Edge.
        :return: A boolean indicating if the specified edge should be added.
        """
        lhs = edge.i
        rhs = edge.j

        if lhs.id == rhs.id:
            return False

        if not PathDetector(self, lhs.id, rhs.id).exists():
            return True

        return False

    def __get_parent_ids__(self, id):
        """
        Gets the parent ids of the specified node id.
        :param id: Node id.
        :return: Array of parent ids.
        """
        node = self.get_bbn_node(id)
        if 'parents' in node.metadata:
            return [n.id for n in node.metadata['parents']]
        return []

    def __notify_listener__(self, change):
        """
        Notifies the JoinTreeListener, if any.
        :param change: ChangeType.
        """
        if self.listener is None:
            return
        if ChangeType.RETRACTION == change:
            self.listener.evidence_retracted(self)
        elif ChangeType.UPDATE == change:
            self.listener.evidence_updated(self)


class PathDetector(object):
    """
    Detects path between two nodes.
    """

    def __init__(self, graph, start, stop):
        """
        Ctor.
        :param graph: Join tree.
        :param start: Start node id.
        :param stop: Stop node id.
        """
        self.graph = graph
        self.start = start
        self.stop = stop
        self.seen = set()

    def exists(self):
        """
        Checks if a path exists.
        :return: True if a path exists, otherwise, false.
        """
        if self.start == self.stop:
            return True
        else:
            return self.__find__(self.start)

    def __find__(self, i):
        """
        Checks if a path exists from the specified node to the stop node.
        :param i: Node id.
        :return: True if a path exists, otherwise, false.
        """
        neighbors = self.graph.get_neighbors(i)
        if self.stop in neighbors:
            return True

        self.seen.add(i)
        for neighbor in neighbors:
            if neighbor not in self.seen and self.__find__(neighbor):
                return True

        return False


class JoinTreeListener(object):
    """
    Interface like class used for listening to a join tree.
    """

    def evidence_retracted(self, join_tree):
        """
        Evidence is retracted.
        :param join_tree: Join tree.
        """
        pass

    def evidence_updated(self, join_tree):
        """
        Evidence is updated.
        :param join_tree: Join tree.
        """
        pass


class EvidenceType(Enum):
    """
    Evidence type.
    """
    VIRTUAL = 1
    FINDING = 2
    OBSERVATION = 3
    UNOBSERVE = 4


class ChangeType(Enum):
    """
    Change type.
    """
    NONE = 1
    UPDATE = 2
    RETRACTION = 3


class EvidenceBuilder(object):
    """
    Evidence builder.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.values = dict()
        self.node = None
        self.type = EvidenceType.OBSERVATION

    def with_node(self, node):
        """
        Adds a BBN node.
        :param node: BBN node.
        :return: Builder.
        """
        self.node = node
        return self

    def with_type(self, type):
        """
        Adds the EvidenceType.
        :param type: EvidenceType.
        :return: Builder.
        """
        self.type = type
        return self

    def with_evidence(self, val, likelihood):
        """
        Adds evidence.
        :param val: Value.
        :param likelihood: Likelihood.
        :return: Builder.
        """
        self.values[val] = likelihood
        return self

    def build(self):
        """
        Builds an evidence.
        :return: Evidence.
        """
        evidence = Evidence(self.node, self.type)
        for k, v in self.values.items():
            evidence.add_value(k, v)
        return evidence


class Evidence(object):
    """
    Evidence.
    """

    def __init__(self, node, type):
        """
        Ctor.
        :param node: BBN node.
        :param type: EvidenceType.
        """
        self.node = node
        self.type = type
        self.values = dict()

    def add_value(self, value, likelihood):
        """
        Adds a value.
        :param value: Value.
        :param likelihood: Likelihood.
        :return: This evidence.
        """
        self.values[value] = likelihood
        return self

    def compare(self, potentials):
        """
        Compares this evidence with previous ones.
        :param potentials: Map of potentials.
        :return: The ChangeType from the comparison.
        """
        that = self.__convert__(potentials)

        that_unobserve = self.__is_unobserved__(that)
        this_unobserve = self.__is_unobserved__(self.values)

        if that_unobserve and this_unobserve:
            return ChangeType.NONE

        that_observe = self.__is_observed__(that)
        this_observe = self.__is_observed__(self.values)

        if that_observe and this_observe:
            s1 = self.__get_observed_value__(that)
            s2 = self.__get_observed_value__(self.values)

            if s1 == s2:
                return ChangeType.NONE
            else:
                return ChangeType.RETRACTION

        return ChangeType.RETRACTION

    @staticmethod
    def __convert__(potentials):
        """
        Converts potentials to a map (dict).
        :param potentials: Potentials.
        :return: Dict where keys are BBN node values and values are likelihoods.
        """
        m = dict()
        for k, v in potentials.items():
            m[k] = v.entries[0].value
        return m

    @staticmethod
    def __is_unobserved__(values):
        """
        Checks if the values represent an unobserved evidence. If all likelihoods are 1.0, then this
        map of values represent unobserved evidence.
        :param values: Map of values, where keys are values and values are likelihoods.
        :return: A boolean indicating if the values represent unobserved evidence.
        """
        count = 0
        for k, v in values.items():
            count += v
        return count == len(values)

    @staticmethod
    def __is_observed__(values):
        """
        Checks if the values represent an observed evidence. If all likelihoods are 0 with exactly
        one of them being 1, then this map of values represent observed evidence.
        :param values: Map of values, where keys are values and values are likelihoods.
        :return: A boolean indicating if the values represent observed evidence.
        """
        one = 0
        zero = 0

        for k, v in values.items():
            if 1.0 == v:
                one += 1
            else:
                zero += 1

        return 1 == one and len(values) - 1 == zero

    @staticmethod
    def __get_observed_value__(values):
        """
        Gets the value that is observed (the value whose likelihood is 1.0).
        :param values: Map of values, where keys are values and values are likelihoods.
        :return: Observed value.
        """
        strs = [k for k in values if 1.0 == values[k]]
        return strs[0]

    def validate(self):
        """
        Validates this evidence.

        * virtual evidence: sum of likelihoods must equal to 1.0
        * finding evidence: all likelihoods must be exactly 1.0 or 0.0.
        * observation evidence: exactly one likelihood is 1.0 and all others must be 0.0.
        """
        for value in self.node.variable.values:
            if value not in self.values:
                self.values[value] = 0.0

        if EvidenceType.VIRTUAL == self.type:
            total = sum([x for x in self.values.values()])
            for value in self.node.variable.values:
                d = self.values[value] / total
                self.values[value] = d
        elif EvidenceType.FINDING == self.type:
            for value in self.node.variable.values:
                d = 1.0 if self.values[value] > 0.0 else 0.0
                self.values[value] = d

            count = sum([x for x in self.values.values()])
            if 0.0 == count:
                for value in self.node.variable.values:
                    self.values[value] = 1.0
        elif EvidenceType.OBSERVATION == self.type:
            tuples = []
            for k, v in self.values.items():
                pair = (k, v)
                tuples.append(pair)
            tuples = sorted(tuples, key=lambda x: (x[1]), reverse=True)
            key = tuples[0][0]
            for value in self.node.variable.values:
                if key == value:
                    self.values[value] = 1.0
                else:
                    self.values[value] = 0.0
        elif EvidenceType.UNOBSERVE == self.type:
            for value in self.node.variable.values:
                self.values[value] = 1.0



from nose import with_setup

from pybbn.graph.dag import BbnUtil
from pybbn.pptc.initializer import Initializer
from pybbn.pptc.moralizer import Moralizer
from pybbn.pptc.potentialinitializer import PotentialInitializer
from pybbn.pptc.propagator import Propagator
from pybbn.pptc.transformer import Transformer
from pybbn.pptc.triangulator import Triangulator


def setup():
    """
    Setup.
    :return: None.
    """
    pass


def teardown():
    """
    Teardown.
    :return: None.
    """
    pass


@with_setup(setup, teardown)
def test_propagator():
    """
    Tests propagation.
    :return: None.
    """
    bbn = BbnUtil.get_huang_graph()
    PotentialInitializer.init(bbn)

    ug = Moralizer.moralize(bbn)
    cliques = Triangulator.triangulate(ug)

    join_tree = Transformer.transform(cliques)

    Initializer.initialize(join_tree)
    Propagator.propagate(join_tree)

    # assert later
    # for clique in join_tree.get_cliques():
    #     potential = join_tree.potentials[clique.id]
    #     print(clique)
    #     print(potential)
    #     total = sum([entry.value for entry in potential.entries])
    #     print('total {}'.format(total))
    # assert 1 == 2



class Variable(object):
    """
    A variable.
    """

    def __init__(self, id, name, values):
        """
        :param id: Numeric identifier. e.g. 0
        :param name: Name. e.g. 'a'
        :param values: Array of values. e.g. ['on', 'off']
        """
        self.id = id
        self.name = name
        self.values = values



from setuptools import setup, find_packages

setup(
    name='pybbn',
    version='0.0.5',
    author='Jee Vang',
    author_email='vangjee@gmail.com',
    packages=find_packages(),
    description='Inference in Bayesian Belief Networks using Probability Propagation in Trees of Clusters (PPTC) and Gibbs sampling',
    url='https://github.com/vangj/py-bbn',
    keywords=['bayesian', 'belief', 'network', 'exact', 'approximate', 'inference', 'junction', 'tree', 'algorithm',
              'pptc', 'dag', 'gibbs', 'sampling', 'multivariate', 'conditional', 'gaussian', 'linear'],
    classifiers=[]
)
